"""Implementation of the optimization methods for the
empirical loss + regularizer
framework, for binary clssification.

The algorithms are primal variable algorithms.

We consider the following losses:
- hinge loss
- ramp loss

We consider the following regularizers:
- l2
- l1
- capped l1
- l0 (approximated by exponent)
"""

from time import time
import numpy
from numpy.random import randn
from numpy import matrix, array, vstack, hstack
from numpy import zeros, ones, transpose, nonzero
from numpy import sign, eye, where
from numpy.linalg import norm

from optwok.sip import SILP, SIQP
from optwok.poly_feature import get_poly_features

class Classifier(object):
    """
    Implementation of the optimization methods for the
    empirical loss + regularizer
    framework, for binary clssification.

    We only consider linear classifiers.
    """
    def __init__(self, reg_param=1.0, kdict=None, verbose=False):
        assert(reg_param > 0.0)
        self.reg_param = reg_param     # Tradeoff regularizer vs loss
        self.regularizer = None
        self.loss = None

        self.kdict = kdict             # kernel
        self.train_ex = array([])      # training examples
        self.y = array([])             # labels
        self.train_time = 0.0          # time used for training
        
        self.w = array([])             # weight
        self.b = array([])             # bias
        self.epsilon = 1e-6            # how small is zero?
        self.verbose = verbose         # how much output

    @property
    def num_examples(self):
        return len(self.y)
    
    @property
    def train_out(self):
        """Output of the classifier on the training set"""
        return self.get_svm_out(self.train_ex)
    
    @property
    def margin(self):
        """The value of the regularizer"""
        return self.regularizer(self.w)

    @property
    def emp_risk(self):
        """A vector of the empirical loss"""
        return self.loss.get_emp_risk(self.y*self.train_out)

    @property
    def primal_obj(self):
        """Directly compute the primal objective function."""
        if self.verbose:
            print 'Margin: %f, Risk: %f' % (self.margin,
                                            self.reg_param*sum(self.emp_risk))
        obj = self.margin + self.reg_param*sum(self.emp_risk)
        return obj[0]

    @property
    def max_weight(self):
        """The maximum absolute value of the weight."""
        return max(abs(self.w))[0]

    @property
    def frac_sv(self):
        """The fraction of examples used as support vectors."""
        return self.loss.get_frac_sv(self.y*self.train_out)

    @property
    def nonzero_feat(self):
        """The features corresponding to non-zero parameters"""
        return numpy.where(abs(self.w)>self.epsilon)[0]
    
    @property
    def frac_feat(self):
        """The fraction of features used in the final classifier,
        i.e. the number of non-zero parameters"""
        return float(len(self.nonzero_feat))/float(len(self.w))

    @property
    def frac_total(self):
        """The fraction of training data used in the final classifier,
        taking both features and support vectors into account."""
        num_sv = float(self.loss.get_num_sv(self.y*self.train_out))
        num_feat = float(len(numpy.where(abs(self.w)>self.epsilon)[0]))
        return (num_sv*num_feat) / (float(self.train_ex.shape[1])*float(len(self.w)))

    @property
    def perf(self):
        """Summary of results"""
        perf = {'max_weight': self.max_weight,
                'nonzero_feat': self.nonzero_feat,
                'frac_feat': self.frac_feat,
                'frac_sv': self.frac_sv,
                'frac_total': self.frac_total,
                'train_time': self.train_time,
                'objective': self.primal_obj,
                }
        return perf

    def get_svm_out(self, X):
        """Compute K*alpha+b"""
        return array(matrix(self.w).T*matrix(X) + self.b).flatten()

    def solve_opt(self):
        """solve_opt should the implemented in child class"""
        pass
    
    def train(self, train_ex, train_lab):
        """Wrapper to convert data to feature matrix,
        and solve the optimization"""

        self.train_ex = compute_features(train_ex, self.kdict)
        self.y = train_lab

        t_start = time()
        self.solve_opt()
        t_end = time()

        # Collect statistics about optimization
        self.train_time = t_end-t_start

        if self.verbose: print '%f support vectors' % self.frac_sv
        if self.verbose: print 'Primal Objective: %f' % self.primal_obj

    def pred(self, pred_ex):
        """Wrapper to do prediction"""
        return self.get_svm_out(compute_features(pred_ex, self.kdict))

    
class ConvexClassifier(Classifier):
    """
    Classifiers which have convex training methods
    """
    def __init__(self, reg_param, kdict, verbose):
        Classifier.__init__(self, reg_param, kdict, verbose)
        self.num_iter = 1

class SVMPrimal(ConvexClassifier):
    """
    Standard l2 regularized SVM, in the primal
    """
    def __init__(self, reg_param, kdict, verbose=False):
        ConvexClassifier.__init__(self, reg_param, kdict, verbose)
        self.loss = HingeLoss()
        self.regularizer = margin_l2

    def solve_opt(self):
        """Solve the SVM QP
        Use cvxopt qp solver which expects the following format:
        min_x   1/2 x'Px + q'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving QP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+d+1
        svm = SIQP(m, self.verbose)
        P = zeros((m, m))
        P[:d,:d] = 2.0*self.reg_param*eye(d)
        q = vstack([zeros((d+1,1)), ones((n,1))])
        svm.set_objective(P,q)
        svm.add_ineq_constraints(hstack([zeros((n,d+1)), -eye(n)]), zeros((n,1)))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n)])
        svm.add_ineq_constraints(G, -ones((n,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

class OneNormSVM(ConvexClassifier):
    """
    One norm SVM
    """    
    def __init__(self, reg_param, kdict, verbose=False):
        ConvexClassifier.__init__(self, reg_param, kdict, verbose)
        self.loss = HingeLoss()
        self.regularizer = margin_l1

    def solve_opt(self):
        """Solve the linear program corresponding to the one norm SVM.
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([zeros((d+1,1)), ones((n,1)),
                                  self.reg_param*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

class DiffConvexClassifier(Classifier):
    """
    Classifiers which have difference of convex training methods
    """
    def __init__(self, reg_param, kdict, verbose=False, rand_init=False):
        Classifier.__init__(self, reg_param, kdict, verbose)
        self.rand_init = rand_init  # Use SVM or random initialization
        self.num_iter = 0
        self.max_iter = 100         # Maximum number of DC iterations
        self.min_diff = 1e-2        # Stopping criteria, change in parameters
        self.all_objective = []     # The objective value for all iterations

    @property
    def perf(self):
        """Summary of results"""
        perf = super(DiffConvexClassifier, self).perf
        perf['num_iter'] = self.num_iter
        perf['all_objective'] = self.all_objective
        return perf

    def _init_param(self, norm):
        """Initialize values of w and b.
        The parameter denotes whether to use l1 or l2 norm.
        """
        assert(norm == 1 or norm==2)
        if self.rand_init:
            self.num_iter = 0
            if self.verbose: print 'Random initialization'
            w = self.reg_param*randn(self.train_ex.shape[0],1)
            b = -randn()
        else:
            self.num_iter = 1
            if self.verbose: print 'SVM initialization'
            if norm == 1:
                svm = OneNormSVM(self.reg_param, self.kdict, self.verbose)
            elif norm == 2:
                svm = SVMPrimal(self.reg_param, self.kdict, self.verbose)
            # train computes the feature mapping explicitly
            svm.train_ex = self.train_ex
            svm.y = self.y
            svm.solve_opt()
            (w, b) = (svm.w, svm.b)

        return (w, b)

    def _subgradient_concave_ramp(self, mu):
        """Compute the subgradient of the concave part of the ramp loss"""
        (clipped, sloped, correct) = self.loss.get_loss_case(self.y*self.train_out)
        #sw = zeros(self.train_ex.shape[0])
        #sb = zeros((1,1))
        #for idx in clipped:
        #    sw += -self.y[idx]*self.train_ex[:,idx]
        #    sb += -self.y[idx]
        sw = numpy.sum(-self.y[clipped]*self.train_ex[:,clipped], axis=1)
        sw.shape = (self.train_ex.shape[0], 1)
        sb = sum(-self.y[clipped])
        sw /= mu
        sb /= mu
        
        return (sw, sb)

    def _subgradient_concave_cap_l1(self, a):
        """Compute the subgradient of the concave part of the capped l1 norm"""
        grad = sign(self.w)
        grad[nonzero(abs(self.w) < a)[0]] = 0.0
        return grad

    def _subgradient_concave_l0(self, a):
        """Compute the subgradient of the concave part of the l0 approximation"""
        return a*sign(self.w)*(1.0 - numpy.exp(-a*abs(self.w)))

class CapOneNormSVM(DiffConvexClassifier):
    """
    SVM with capped one norm regularizer.
    """
    def __init__(self, reg_param, kdict, a=1.0, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.a = a                      # The value of the cap
        self.loss = HingeLoss()
        self.regularizer = margin_cap_l1
        
    @property
    def margin(self):
        """The value of the regularizer, specialized because of parameter a"""
        return self.regularizer(self.w, self.a)

    def _solve_linprog(self, u):
        """Solve the linear program corresponding to the convex
        part of the capped one norm SVM.
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([-self.reg_param*u, zeros((1,1)), ones((n,1)),
                                  self.reg_param*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(1)
        self.all_objective.append(self.primal_obj)
        u = self._subgradient_concave_cap_l1(self.a)
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_linprog(u)
            old_u = u
            u = self._subgradient_concave_cap_l1(self.a)
            diff_param = sum(abs(old_u-u))
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)


class ZeroNormSVM(DiffConvexClassifier):
    """
    SVM with zero norm regularizer. Uses exponential approximation.
    """
    def __init__(self, reg_param, kdict, a=4.0, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.a = a                      # The exponent in approximation
        self.loss = HingeLoss()
        self.regularizer = margin_l0
        
    @property
    def margin(self):
        """The value of the regularizer, specialized because of parameter a"""
        return self.regularizer(self.w, self.a)

    def _solve_linprog(self, u):
        """Solve the linear program corresponding to the convex
        part of the zero norm SVM.
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([-self.reg_param*u, zeros((1,1)), ones((n,1)),
                                  self.reg_param*self.a*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(1)
        self.all_objective.append(self.primal_obj)
        u = self._subgradient_concave_l0(self.a)
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_linprog(u)
            old_u = u
            u = self._subgradient_concave_l0(self.a)
            diff_param = sum(abs(old_u-u))
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)

class RLSVM2(DiffConvexClassifier):
    """
    Ramp Loss SVM with l2 regularizer.
    """
    def __init__(self, reg_param, kdict, mu, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.mu = mu                   # the slope of the ramp
        self.loss = RampLoss(mu)
        self.regularizer = margin_l2
        
    def _solve_quadprog(self, sw, sb):
        """Solve the quadratic program given by
        linearizing the concave part

        Use cvxopt qp solver which expects the following format:
        min_x   1/2 x'Px + q'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving QP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+d+1
        svm = SIQP(m, self.verbose)
        P = zeros((m,m))
        P[:d,:d] = 2.0*self.reg_param*eye(d)
        q = vstack([-sw, -sb, ones((n,1))/self.mu])
        svm.set_objective(P, q)
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n)]), zeros((n,1)))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n)])
        svm.add_ineq_constraints(G, -self.mu*ones((n,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(2)
        self.all_objective.append(self.primal_obj)
        (sw, sb) = self._subgradient_concave_ramp(self.mu)
        #print sw, sb
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_quadprog(sw, sb)
            #print 'weights'
            #print self.w, self.b
            (old_sw, old_sb) = (sw, sb)
            (sw, sb) = self._subgradient_concave_ramp(self.mu)
            #print sw, sb
            diff_param = sum(abs(old_sw-sw)) + abs(old_sb-sb)
            #print diff_param
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)

class RLSVM1(DiffConvexClassifier):
    """
    Ramp Loss SVM with l1 regularizer.
    """
    def __init__(self, reg_param, kdict, mu, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.mu = mu                   # the slope of the ramp
        self.loss = RampLoss(mu)
        self.regularizer = margin_l1
        
    def _solve_linprog(self, sw, sb):
        """Solve the linear program corresponding to the convex part
        of the ramp loss SVM with l1 penalty.
        
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([-sw, -sb, ones((n,1))/self.mu,
                                  self.reg_param*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -self.mu*ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM with one norm penalty"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(1)
        self.all_objective.append(self.primal_obj)
        (sw, sb) = self._subgradient_concave_ramp(self.mu)
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_linprog(sw, sb)
            (old_sw, old_sb) = (sw, sb)
            (sw, sb) = self._subgradient_concave_ramp(self.mu)
            diff_param = sum(abs(old_sw-sw)) + abs(old_sb-sb)
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)

class RLSVMC1(DiffConvexClassifier):
    """
    Ramp Loss SVM with capped l1 regularizer.
    """
    def __init__(self, reg_param, kdict, mu, a=1.0, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.mu = mu                   # the slope of the ramp
        self.a = a                     # The value of the cap
        self.loss = RampLoss(mu)
        self.regularizer = margin_cap_l1
        
    @property
    def margin(self):
        """The value of the regularizer, specialized because of parameter a"""
        return self.regularizer(self.w, self.a)

    def _solve_linprog(self, sw, sb, u):
        """Solve the linear program corresponding to the convex part
        of the ramp loss SVM with capped l1 penalty.
        
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([-sw-self.reg_param*u, -sb, ones((n,1))/self.mu,
                                  self.reg_param*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -self.mu*ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM with capped l1 regularizer"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(1)
        self.all_objective.append(self.primal_obj)
        (sw, sb) = self._subgradient_concave_ramp(self.mu)
        u = self._subgradient_concave_cap_l1(self.a)
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_linprog(sw, sb, u)
            (old_sw, old_sb, old_u) = (sw, sb, u)
            (sw, sb) = self._subgradient_concave_ramp(self.mu)
            u = self._subgradient_concave_cap_l1(self.a)
            diff_param = sum(abs(old_sw-sw)) + abs(old_sb-sb) + sum(abs(old_u-u))
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)


class RLSVM0(DiffConvexClassifier):
    """
    Ramp Loss SVM with l0 regularizer.
    """
    def __init__(self, reg_param, kdict, mu, a=4.0, verbose=False, rand_init=False):
        DiffConvexClassifier.__init__(self, reg_param, kdict, verbose, rand_init)
        self.mu = mu                   # the slope of the ramp
        self.a = a                     # The exponent in approximation
        self.loss = RampLoss(mu)
        self.regularizer = margin_l0
        
    @property
    def margin(self):
        """The value of the regularizer, specialized because of parameter a"""
        return self.regularizer(self.w, self.a)

    def _solve_linprog(self, sw, sb, u):
        """Solve the linear program corresponding to the convex part
        of the ramp loss SVM with l0 penalty.
        
        Use cvxopt lp solver which expects the following format:
        min_x   c'x
        st      Gx <= h
                Ax = b

        We organise the variables of the SVM as (w,b,xi,eta).
        """
        yvec = self.y.copy()
        yvec.shape = (len(self.y), 1)

        if self.verbose: print 'Solving LP'
        n = len(self.y)
        d = self.train_ex.shape[0]
        m = n+2*d+1
        svm = SILP(m, self.verbose)
        svm.set_objective(vstack([-sw-self.reg_param*u, -sb, ones((n,1))/self.mu,
                                  self.reg_param*self.a*ones((d,1))]))
        G = hstack([-yvec*transpose(self.train_ex), -yvec, -eye(n), zeros((n,d))])
        svm.add_ineq_constraints(G, -self.mu*ones((n,1)))
        svm.add_ineq_constraints(hstack([zeros((n,d+1)),-eye(n),zeros((n,d))]), zeros((n,1)))
        svm.add_ineq_constraints(hstack([zeros((d,d+1)),zeros((d,n)),-eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.add_ineq_constraints(hstack([-eye(d), zeros((d,n+1)), -eye(d)]), zeros((d,1)))
        svm.solve()

        if svm.optimal:
            self.w = svm.solution[:d]
            self.w.shape = (d,1)
            self.b = svm.solution[d]

    def solve_opt(self):
        """Use a DCA to solve the ramp loss SVM with zero norm (approx) regularizer"""
        if self.verbose: print 'Solving DCA'
        (self.w, self.b) = self._init_param(1)
        self.all_objective.append(self.primal_obj)
        (sw, sb) = self._subgradient_concave_ramp(self.mu)
        u = self._subgradient_concave_l0(self.a)
        diff_param = numpy.inf
        while (self.num_iter < self.max_iter)\
              and (diff_param > self.min_diff):
            self._solve_linprog(sw, sb, u)
            (old_sw, old_sb, old_u) = (sw, sb, u)
            (sw, sb) = self._subgradient_concave_ramp(self.mu)
            u = self._subgradient_concave_l0(self.a)
            diff_param = sum(abs(old_sw-sw)) + abs(old_sb-sb) + sum(abs(old_u-u))
            self.num_iter += 1
            self.all_objective.append(self.primal_obj)

class Loss(object):
    """A loss function, for classification"""
    epsilon = 0.01

class HingeLoss(Loss):
    """
    Standard SVM loss
    """
    def get_emp_risk(self, t):
        """Compute the empirical risk, for each example."""
        (sloped, correct) = self.get_loss_case(t)
        emp_risk = zeros(len(t))
        emp_risk[sloped] = 1.0 - t[sloped]
        return emp_risk

    def get_num_sv(self, t):
        """Estimate the number of support vectors"""
        (sloped, correct) = self.get_loss_case(t)
        return len(sloped)

    def get_frac_sv(self, t):
        """Estimate the fraction of support vectors"""
        num_sv = self.get_num_sv(t)
        frac_sv = float(num_sv)/float(len(t))
        return frac_sv

    def get_loss_case(self, t):
        """Detect whether we are on the slope,
        or classified correctly (larger than margin)."""
        sloped = numpy.where(t <= 1.0+self.epsilon)[0]
        correct = numpy.where(t > 1.0+self.epsilon)[0]
        return (sloped, correct)

class RampLoss(Loss):
    """
    Truncated hinge loss
    """
    def __init__(self, mu):
        """The parameter mu controls the slope between [0,2]"""
        assert((mu > 0.0) and (mu <= 1.0))
        self.mu = mu
        
    def get_emp_risk(self, t):
        """Compute the empirical risk, for each example."""
        (clipped, sloped, correct) = self.get_loss_case(t)
        emp_risk = zeros(len(t))
        emp_risk[clipped] = self.mu + 1.0
        emp_risk[sloped] = 1.0 - (t[sloped]/self.mu)
        return emp_risk

    def get_num_sv(self, t):
        """Estimate the number of support vectors"""
        (clipped, sloped, correct) = self.get_loss_case(t)
        return len(sloped)

    def get_frac_sv(self, t):
        """Estimate the fraction of support vectors"""
        num_sv = self.get_num_sv(t)
        frac_sv = float(num_sv)/float(len(t))
        return frac_sv

    def get_loss_case(self, t):
        """Detect whether we are at saturation, on the slope,
        or classified correctly (larger than margin)."""
        clipped = numpy.where(t<-(self.mu*self.mu)-self.epsilon)[0]
        lb = numpy.where(t>=-(self.mu*self.mu)-self.epsilon, 1, 0)
        ub = numpy.where(t<=self.mu+self.epsilon, 1, 0)
        sloped = numpy.where((lb+ub)>1)[0]
        correct = numpy.where(t>self.mu+self.epsilon)[0]
        return (clipped, sloped, correct)

class SlantLoss(RampLoss):
    """
    Hinge loss with double kink
    """
    def get_emp_risk(self, t):
        """Compute the empirical risk, for each example."""
        (clipped, sloped, correct) = self.get_loss_case(t)
        emp_risk = zeros(len(t))
        emp_risk[clipped] = 0.5*(3.0-t[clipped])
        emp_risk[sloped] = 1.0 - t[sloped]

        return emp_risk


def margin_l2(w):
    """The l2 norm of a vector"""
    margin = matrix(w).T*matrix(w)
    return margin[0,0]

def margin_l1(w):
    """The l1 norm of a vector"""
    return sum(abs(w))

def margin_cap_l1(w, a):
    """The capped l1 norm of a vector"""
    assert(a > 0.0)
    cap_l1 = abs(w)
    cap_l1[nonzero(cap_l1 > a)[0]] = a
    return sum(cap_l1)

def margin_l0(w, a):
    """The l0 norm of a vector, i.e. the number of non-zeros.
    We approximate by an exponential.
    """
    assert(a > 0.0)
    return sum(1.0 - numpy.exp(-a*abs(w)))

def compute_features(train_ex, kdict):
    """Explicitly compute the feature space of the kernel"""
    if kdict['name'] == 'linear':
        return train_ex
    elif kdict['name'] == 'poly':
        if kdict['inhomog'] == True:
            train_ex = vstack([train_ex, ones((1,train_ex.shape[1]))])
        phi = get_poly_features(train_ex, kdict['param'])
        return phi
    else:
        print 'Explicit feature construction for kernel %s not implemented' % kdict['name']


