"""Optimize over the choice of kernel.
For learning the kernel over labels.
"""

from time import time
from numpy import zeros, ones, eye, sign
from numpy import matrix, array, diag, trace
from numpy import vstack, unique, maximum, argmax
from numpy.linalg import eig, norm, solve, lstsq, inv
from numpy import diag, sqrt, divide, logspace, kron
from numpy import min, max, transpose, real
from scipy.linalg.matfuncs import sqrtm
from matplotlib.pyplot import matshow, title, colorbar, savefig, show, gray
from matplotlib.colors import LinearSegmentedColormap
from cvxopt import matrix as cm
from cvxopt.solvers import sdp
from optwok.multiclass import StructPredictor
from optwok.multiclass import idx2binvec
from optwok.kernel import normalize_unit_diag, JointKernel
from optwok.io_pickle import load, save
from slycot.synthesis import sb04qd as dlyap

class KernelOptBase(object):
    """Base class for learning the kernel on labels"""
    def __init__(self, reg_param=1.0, verbose=False):
        assert(reg_param > 0.0)
        self.reg_param = reg_param     # Tradeoff regularizer vs loss
        self.all_reg_param = logspace(-4,4,30)
        self.verbose = verbose         # how much output
        self.Ky = array([])            # kernel on labels
        self.all_Ky = None             # label kernel for whole reg path

    def train(self, train_ex, train_lab):
        raise NotImplementedError

    def predict(self, pred_ex):
        raise NotImplementedError

    def save_Ky(self, dataname, work_dir='.'):
        """Save the learned label kernel,
        as pickle and pdf image"""
        if self.all_Ky is None:
            save('%s/%s_Ky.pkl' % (work_dir, dataname), self.Ky)
        else:
            save('%s/%s_Ky.pkl' % (work_dir, dataname), self.all_Ky)
        show_kernel(self.Ky)
        title(dataname)
        savefig('%s/%s_Ky.pdf' % (work_dir, dataname))
            
class KernelOptLS(KernelOptBase):
    """Optimize the kernel on the labels,
    with the least squares loss, and also Frobenious norm regularizer.
    Uses slycot for solving the sylvester equation"""
    def __init__(self, reg_param=1.0, kernel=None, verbose=False):
        KernelOptBase.__init__(self, reg_param, verbose)
        self.max_iter = 1000           # maximum number of iterations
        self.rel_tol = 0.01            # relative stopping tolerance
        self.near_zero = 1e-5          # initialize a positive definite matrix

        self.train_ex = array([])      # training examples
        self.train_lab = array([])     # labels in original format
        self.y = array([])             # labels
        self.num_classes = 0           # number of classes
        self.num_train = 0             # number of training examples

        #self.objective = []            # objective value at each iteration
        self.Kx = kernel               # kernel on examples
        self.alpha = array([])         # support vector coefficients
        
    def train(self, train_ex, train_lab, init_Ky=None):
        """Wrapper to convert data to feature matrix,
        and solve the optimization"""
        self.train_ex = train_ex
        self.train_lab = train_lab
        self._set_dimensions(train_lab)
        self.Kx.compute(train_ex)
        if init_Ky is not None:
            self.Ky = init_Ky
        else:
            #self.Ky = zeros((self.num_classes,self.num_classes))
            self.Ky = self.near_zero*eye(self.num_classes)
        # shrink Ky towards this matrix
        #bias = matrix(ones((self.num_classes,self.num_classes)))
        #bias = matrix(eye(self.num_classes))

        residual = norm(self.y)
        min_tol = self.rel_tol * residual
        num_iter = 0

        print 'Learning the kernel by coordinate descent'
        while residual > min_tol and num_iter < self.max_iter:
            num_iter += 1
            self.alpha = dlyap(self.num_train, self.num_classes,
                          self.Kx.kmat, self.Ky/self.reg_param,
                          self.y/self.reg_param)
            #self.objective.append((trace(matrix(self.alpha).T*matrix(self.y))+
            #                            norm(self.Ky)**2)/2.0)
            X = matrix(self.Kx.kmat)*matrix(self.alpha)
            #P = 0.5*X.T*matrix(self.alpha)-matrix(self.Ky)+bias
            P = 0.5*X.T*matrix(self.alpha)-matrix(self.Ky)
            P = (P+P.T)/2.0
            Q = matrix(solve(X.T*X+self.reg_param*eye(self.num_classes), P))
            #(Q, res, xrank, s) = lstsq(X.T*X+self.reg_param*eye(self.num_classes), P)
            Q = (Q+Q.T)/2.0
            self.Ky += self.reg_param*Q
            yhat = X*matrix(self.Ky)
            residual = norm(yhat+self.reg_param*self.alpha-self.y)
            if self.verbose:
                print 'Iteration: %d: residual = %f' % (num_iter, residual)

        #self.normalize_unit_diag()
        print 'Number of iterations: %d' % num_iter
        if self.verbose:
            print 'Kernel on labels'
            print self.Ky

    def train_reg_path(self, train_ex, train_lab, all_reg_param=None):
        """Solve for the whole regularization path"""
        if all_reg_param is not None:
            self.all_reg_param = all_reg_param
        self._set_dimensions(train_lab)
        self.all_Ky = zeros((self.num_classes, self.num_classes, len(self.all_reg_param)))
        self.Ky = zeros((self.num_classes, self.num_classes))
        for (idx_reg, reg) in reversed(list(enumerate(self.all_reg_param))):
            self.reg_param = reg
            self.train(train_ex, train_lab, self.Ky)
            self.all_Ky[:,:,idx_reg] = self.Ky.copy()
        # Use the middle of the regularization range as default
        self.Ky = self.all_Ky[:,:,round(self.all_Ky.shape[2]/2)]

    def infer_sv(self, Ky=None):
        """Compute the support vector coefficients for a given Ky"""
        if Ky is not None:
            self.Ky = Ky
        self.alpha = dlyap(self.num_train, self.num_classes,
                           self.Kx.kmat, self.Ky/self.reg_param,
                           self.y/self.reg_param)

    def predict(self, pred_ex, get_max=True):
        """Wrapper to do prediction.
        returns argmax of scores"""
        Ktest = self.Kx.compute_test(pred_ex)
        scores = array((matrix(self.Ky)*matrix(self.alpha).T*matrix(Ktest)).T)
        if get_max:
            return argmax(scores, axis=1)
        else:
            return transpose(scores)

    def predict_reg_path(self, pred_ex, ident_Ky=False):
        """Compute prediction for each value in regularization path.
        Use learned Ky and retrain to get the corresponding
        support vector coefficients.
        """
        all_preds = zeros((pred_ex.shape[1],len(self.all_reg_param)))
        for (idx_reg, reg) in reversed(list(enumerate(self.all_reg_param))):
            print 'Predicting with reg param = %e' % reg
            self.reg_param = reg
            if ident_Ky:
                self.infer_sv(eye(self.num_classes))
            else:
                self.infer_sv(self.all_Ky[:,:,idx_reg])
            all_preds[:,idx_reg] = self.predict(pred_ex)
        return all_preds
        
    def normalize_unit_diag(self):
        """Normalize Ky to unit diagonal, and appropriately scale alpha"""
        diag_Ky = diag(self.Ky)
        diag_Ky.shape = (self.num_classes, 1)        
        self.Ky = divide(self.Ky, sqrt(matrix(diag_Ky)*matrix(diag_Ky).T))
        diag_Ky.shape = (1, self.num_classes)
        diag_Ky = sqrt(diag_Ky)
        self.alpha = self.alpha*diag_Ky

    def _set_dimensions(self, train_lab):
        """Infer the number of examples and classes"""
        if len(train_lab.shape) == 1:
            self.y = idx2binvec(train_lab)
            self.num_train = len(train_lab)
            self.num_classes = len(unique(train_lab))
            assert(self.y.shape == (self.num_train, self.num_classes))
        else:
            self.y = train_lab.T
            (self.num_classes, self.num_train) = train_lab.shape
        
class KernelOptLSGP(KernelOptLS):
    """Optimize the kernel on the labels,
    with the least squares loss, and also Frobenious norm regularizer.
    Uses slycot for solving the sylvester equation.

    Multiple output regression, also estimate predictive variance.
    """
    def __init__(self, reg_param=1.0, kernel=None, verbose=False):
        KernelOptLS.__init__(self, reg_param, kernel(), verbose)
        self.test_kernel = JointKernel(kernel(), self.Ky)

    def predict(self, pred_ex):
        """Wrapper to do prediction and also estimate predictive variance."""
        num_test = pred_ex.shape[1]
        Ktest = self.Kx.compute_test(pred_ex)
        scores = array((matrix(self.Ky)*matrix(self.alpha).T*matrix(Ktest)).T)

        self.test_kernel.Ky = self.Ky
        self.test_kernel.compute(pred_ex)
        A = matrix(inv(kron(self.Kx.kmat,self.Ky)\
                       +self.reg_param*eye(self.num_train*self.num_classes)))
        k_star = matrix(kron(Ktest, self.Ky))
        pred_var = diag(self.test_kernel.kmat - k_star.T*A*k_star)
        pred_var.shape = (num_test, self.num_classes)
        return scores.T, pred_var.T

    def predict_reg_path(self, pred_ex, ident_Ky=False):
        """Compute prediction for each value in regularization path.
        Use learned Ky and retrain to get the corresponding
        support vector coefficients.
        """
        all_preds = zeros((self.num_classes, pred_ex.shape[1], len(self.all_reg_param)))
        for (idx_reg, reg) in reversed(list(enumerate(self.all_reg_param))):
            print 'Predicting with reg param = %e' % reg
            self.reg_param = reg
            if ident_Ky:
                self.infer_sv(eye(self.num_classes))
            else:
                self.infer_sv(self.all_Ky[:,:,idx_reg])
            all_preds[:,:,idx_reg] = self.predict(pred_ex)
        return all_preds
        
class KernelOpt(object):
    """Learn the kernel on the labels
    Uses semidefinite programming with cvxopt
    """
    def __init__(self, reg_param=1.0, kernel=None, verbose=False):
        self.train_time = 0.0          # time used for training
        self.max_iter = 3              # maximum number of iterations
        self.verbose = verbose         # how much output

        # Pointer to the classifier object
        self.classifier = StructPredictor(reg_param, kernel, verbose)

    def solve_opt(self, alpha):
        """Solve the optimization problem using an SDP.
        max tr(alpha' Kx alpha Ky)
        st  diag(Ky) = I
            Ky > 0

        cvxopt takes the problem as:
        sdp(c[, Gl, hl[, Gs, hs[, A, b]]])

        minimize     c'x
        subject to   Gl x + sl = hl
                     sl >= 0
                     Ax = b
                     Gk x + vec(sk) = vec(hk)
                     sk > 0

        Gl and hl define the inequality constraints
        Gs and hs are lists, containing Gk and hk respectively, representing the LMIs.
        The inequality > mean positive semidefinite.

        We consider the dual form:
        max -hl'zl - sum_k tr(hk zk) - b'y
        st  Gl'zl + sum_k Gk'vec(zk) + A'y + c = 0
        zl >= 0
        zk > 0
        """
        n = alpha.shape[1]  # the number of classes
        m = n**2            # the size of Ky

        # TODO: How do we regularize?
        h = [cm(-array(normalize_unit_diag(matrix(alpha).T*self.classifier.Kx*matrix(alpha))))]
        #h = [cm(-array(matrix(alpha).T*self.classifier.Kx*matrix(alpha)))]
        #print -h[0]
        A = zeros((m,n))
        for icol in range(n):
            A[(n+1)*icol,icol] = 1.0
        G = [cm(A)]
        c = cm(-ones((n,1)))

        print 'Solving SDP'
        sol = sdp(c, Gs=G, hs=h, solver='dsdp')
        Ky = array(sol['zs'][0])
        Ky = Ky + Ky.T - diag(Ky.diagonal())
        #Ky = maximum(Ky, 0.0)
        #print trace(cm(Ky)*h[0])
        return Ky

    def solve_opt_abspos(self, alpha):
        """Solve the optimization problem using an SDP.
        Also require that entries in Ky are positive.

        cvxopt takes the problem as:
        sdp(c[, Gl, hl[, Gs, hs[, A, b]]])

        minimize     c'x
        subject to   Gl x + sl = hl
                     sl >= 0
                     Ax = b
                     Gk x + vec(sk) = vec(hk)
                     sk > 0

        Gl and hl define the inequality constraints
        Gs and hs are lists, containing Gk and hk respectively, representing the LMIs.
        The inequality > mean positive semidefinite.

        We consider the dual form:
        max -hl'zl - sum_k tr(hk zk) - b'y
        st  Gl'zl + sum_k Gk'vec(zk) + A'y + c = 0
        zl >= 0
        zk > 0
        """
        n = alpha.shape[1]  # the number of classes
        m = n**2            # the size of Ky
        nm = n+m            # The number of variables in the SDP

        # TODO: How do we regularize?
        #h = [cm(-array(normalize_unit_diag(matrix(alpha).T*self.classifier.Kx*matrix(alpha))))]
        #h = [cm(-array(matrix(alpha).T*self.classifier.Kx*matrix(alpha)))]
        F = zeros((nm, nm))
        F[:n,:n] = -array(normalize_unit_diag(matrix(alpha).T*self.classifier.Kx*matrix(alpha)))
        h = [cm(F)]
        #print -h[0]
        Fk = zeros((nm**2,nm))
        for icol in range(n):
            Fk[(nm+1)*icol,icol] = 1.0
        for irow in range(n):
            for icol in range(n):
                Fk[nm*irow+icol,n+n*irow+icol] = 1.0
                Fk[n*nm+n+(nm+1)*(n*irow+icol),n+n*irow+icol] = -1.0
        G = [cm(Fk)]
        c = cm(vstack([-ones((n,1)), zeros((m,1))]))
        print 'Solving SDP'
        sol = sdp(c, Gs=G, hs=h, solver='dsdp')
        Ky = array(sol['zs'][0][:n,:n])
        Ky = Ky + Ky.T - diag(Ky.diagonal())
        #Ky = maximum(Ky, 0.0)
        #print trace(cm(Ky)*h[0])
        return Ky

    def train(self, train_ex, train_lab):
        """Wrapper to convert data to feature matrix,
        and solve the optimization"""
        self.classifier.train(train_ex, train_lab)
        for ix in range(self.max_iter):
            print 'Iteration: %d' % ix
            self.classifier.Ky = self.solve_opt(self.classifier.alpha)
            self.classifier.solve_opt()
            #print self.classifier.Ky
            #print eig(self.classifier.Ky)[0]

    def predict(self, pred_ex):
        """Wrapper to do prediction."""
        return self.classifier.predict(pred_ex)

def show_kernel(kmat):
    """Use matshow with gray colormap"""
    matshow(kmat)
    gray()
    colorbar()


def show_kernel_dev(kmat):
    """Use matshow with a nice colormap"""
    minval = min(kmat.flatten())
    maxval = max(kmat.flatten())
    absval = max([abs(minval),abs(maxval)])
    near_zero = 0.1*absval
    if minval < 0.0:
        zeroval = -minval/(2.0*absval)
    else:
        zeroval = 0.0
    small_neg = zeroval-near_zero
    small_pos = min([1.0, zeroval+near_zero])
    if small_neg >= 0.0:
        blueval = (small_neg-minval)/(small_neg + absval)
        assert(blueval>=0.0 and blueval<=1.0)
        print blueval
        cdict = {'red':  ((0.0, blueval, blueval),
                          (small_neg, 1.0, 1.0),
                          (zeroval, 1.0, 1.0),
                          (small_pos, 1.0, 1.0),
                          (1.0, 1.0, 1.0)),

                 'green': ((0.0, 0.8-blueval, 0.8-blueval),
                           (small_neg, 1.0, 1.0),
                           (zeroval, 1.0, 1.0),
                           (small_pos, 1.0, 1.0),
                           (1.0, 0.2, 0.2)),

                 'blue':  ((0.0, 1-blueval, 1-blueval),
                           (small_neg, 1.0, 1.0),
                           (zeroval, 1.0, 1.0),
                           (small_pos, 1.0, 1.0),
                           (1.0, 0.0, 0.0))
                 }
    else:
        small_neg = 0.0
        cdict = {'red':  ((0.0, 1.0, 1.0),
                          (small_neg, 1.0, 1.0),
                          (zeroval, 1.0, 1.0),
                          (small_pos, 1.0, 1.0),
                          (1.0, 1.0, 1.0)),

                 'green': ((0.0, 1.0, 1.0),
                           (small_neg, 1.0, 1.0),
                           (zeroval, 1.0, 1.0),
                           (small_pos, 1.0, 1.0),
                           (1.0, 0.2, 0.2)),

                 'blue':  ((0.0, 1.0, 1.0),
                           (small_neg, 1.0, 1.0),
                           (zeroval, 1.0, 1.0),
                           (small_pos, 1.0, 1.0),
                           (1.0, 0.2, 0.2))
                 }
    blue_red = LinearSegmentedColormap('BlueRedZero', cdict)
    matshow(kmat, cmap=blue_red)
    colorbar()


def test_kernelopt():
    """Small test for class KernelOpt"""
    from numpy.random import randn
    X = randn(10,50)
    Kx = array(matrix(X).T*matrix(X))
    alpha = randn(50,4)
    print normalize_unit_diag(matrix(alpha).T*matrix(Kx)*matrix(alpha))
    kern_mach = KernelOpt()
    kern_mach.classifier.Kx = Kx
    print kern_mach.solve_opt(alpha)

def _create_test_data(mix_list=[[0,1],[2],[3],[4]],
                      num_point=[100, 50, 50, 50],
                      num_feat=20, edge_length=3):
    """Create an multiclass dataset with mixed classes"""
    from optwok.simdata import multi_cloudgen_mix
    
    num_blobs = len(mix_list)
    train_data = multi_cloudgen_mix(num_point, num_feat, num_blobs, edge_length, mix_list)
    (metadata, examples, labels) = train_data
    return (metadata, examples, labels)

def test_kerneloptls():
    """Small test for class KernelOptLS"""
    from optwok.kernel import LinearKernel

    (metadata, examples, labels) = _create_test_data()
    print metadata
    svm = KernelOptLS(reg_param=1000.0, kernel=LinearKernel(), verbose=True)
    svm.train(examples, labels)
    train_out = svm.predict(examples)
    print vstack([train_out, labels])
    show_kernel(svm.Ky)
    savefig('testKy.pdf')

def test_gp(mix_list=[[0,1],[2]]):
    """Small test for class KernelOptLS,
    for Gaussian process regression"""
    from optwok.kernel import LinearKernel
    from optwok.simdata import gp_gen
    from optwok.evaluation import rmse

    (examples, labels) = gp_gen(100, mix_list)
    svm = KernelOptLSGP(reg_param=10.0, kernel=LinearKernel(), verbose=True)
    svm.train(examples, labels)
    train_out = svm.predict(examples)
    #print vstack([train_out, labels])
    print 'Error:'
    print rmse(train_out, labels)
    show_kernel(svm.Ky)
    
def test_reg_path_Ky():
    """The regularization path"""
    from optwok.kernel import LinearKernel

    (metadata, examples, labels) = _create_test_data()
    print metadata
    svm = KernelOptLS(kernel=LinearKernel(), verbose=True)
    svm.train_reg_path(examples, labels)
    for ix in range(svm.all_Ky.shape[2]):
        show_kernel(normalize_unit_diag(svm.all_Ky[:,:,ix]))
        title(ix)

def test_reg_path_acc():
    """Check accuracy for whole regularization path"""
    from optwok.kernel import LinearKernel
    from optwok.evaluation import accuracy
    from pylab import semilogx, xlabel, ylabel, legend
    
    (metadata, examples, labels) = _create_test_data()
    print metadata
    svm = KernelOptLS(kernel=LinearKernel(), verbose=True)
    svm.train_reg_path(examples, labels)
    (metadata, examples, labels) = _create_test_data()
    all_preds = svm.predict_reg_path(examples)
    
    all_acc = zeros(len(svm.all_reg_param))
    for ix in range(all_preds.shape[1]):
        all_acc[ix] = accuracy(all_preds[:,ix], labels)
    svm.reg_param = 1.0
    svm.infer_sv(eye(svm.num_classes))
    diag_acc = accuracy(svm.predict(examples), labels)
    
    semilogx(svm.all_reg_param, all_acc, hold=True, label='learn Ky')
    semilogx(svm.all_reg_param, all_acc, 'rx')
    semilogx([svm.all_reg_param[0],svm.all_reg_param[-1]], [diag_acc, diag_acc], 'g-', label='identity')
    xlabel('regularization parameter')
    ylabel('accuracy (higher is better)')
    legend()
    
if __name__ == '__main__':
    #test_kerneloptls()
    #test_reg_path_Ky()
    #test_reg_path_acc()
    test_gp()
    show()
